/*----------------------------------------------------------------------
  File    : sphere.c
  Contents: compute grid points on a sphere
  Author  : Christian Borgelt
  History : 10.12.2000 file created
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <assert.h>

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI        3.14159265358979323846  /* \pi */
#endif

/* --- error codes --- */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* file open failed */
#define E_FREAD     (-3)        /* file read failed */
#define E_FWRITE    (-4)        /* file write failed */
#define E_OPTION    (-5)        /* unknown option */
#define E_OPTARG    (-6)        /* missing option argument */
#define E_ARGCNT    (-7)        /* wrong number of arguments */
#define E_SECCNT    (-8)        /* invalid number of sections */
#define E_UNKNOWN   (-9)        /* unknown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {    /* error messages */
  /* E_NONE     0 */  "no error",
  /* E_NOMEM   -1 */  "not enough memory",
  /* E_FOPEN   -2 */  "cannot open file %s",
  /* E_FREAD   -3 */  "read error on file %s",
  /* E_FWRITE  -4 */  "write error on file %s",
  /* E_OPTION  -5 */  "unknown option -%c",
  /* E_OPTARG  -6 */  "missing option argument",
  /* E_ARGCNT  -7 */  "wrong number of arguments",
  /* E_SECCNT  -8 */  "invalid number of sections",
  /* E_UNKNOWN -9 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static const char *prgname;     /* program name for error messages */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print an error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print the error message */
    va_end(args);               /* end argument evaluation */
  }
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  int    i, k = 0;              /* loop variables, counter */
  char   *s;                    /* to traverse the options */
  char   **optarg = NULL;       /* option argument */
  char   *fname   = NULL;       /* name of the output file */
  int    cnt      = 9;          /* num. of points per quarter circle */
  double phi, theta;            /* angles of a point */
  double d;                     /* increment for these angles */
  double st;                    /* sine of theta */
  double x, y, z;               /* temporary buffers */
  FILE   *out;                  /* output file */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1)                 /* if arguments are given */
    printf("%s - compute grid points on a sphere", argv[0]);
  else {                        /* if no argument is given */
    printf("usage: %s [options] outfile\n", argv[0]);
    printf("compute grid points on a sphere\n");
    printf("-s#      number of points per quarter circle "
                    "(default: %d)\n", cnt);
    printf("outfile  file to write points to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 's': cnt = (int)strtol(s, &s, 10); break;
          default : error(E_OPTION, *--s);        break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fname = s;       break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg)  error(E_OPTARG); /* check option argument */
  if (k != 1)  error(E_ARGCNT); /* check number of arguments */
  if (cnt < 1) error(E_SECCNT); /* check the number of points */
  printf("\n");                 /* terminate the startup message */

  /* --- compute and write points --- */
  printf("writing %s ... ", fname); fflush(stdout);
  out = fopen(fname, "w");      /* open the output file */
  if (!out) error(E_FOPEN, fname);
  d     = M_PI/2/cnt;           /* compute angle increment */
  fprintf(out, "%g %g %g\n", 0.0, 0.0, -1.0);
  theta = M_PI;                 /* start at the south pole and */
  for (i = 2*cnt; --i > 0; ) {  /* traverse the parallels of latitude */
    theta -= d;                 /* decrement the polar angle */
    st = sin(theta);            /* compute sine and cosine */
    z  = cos(theta);            /* and the z-coordinate */
    if (fabs(z) <    1e-15) z =  0;  /* avoid roundoff errors */
    else if (z  >  1-1e-15) z =  1;  /* by adapting z */
    else if (z  < -1+1e-15) z = -1;  /* to 0, 1, or -1 */
    phi = 0;                    /* start in the east and */
    for (k = 4*cnt; --k >= 0;){ /* traverse the meridians */
      x = st*cos(phi);          /* compute x avoiding roundoff errors */
      if (fabs(x) <    1e-15) x =  0;
      else if (x  >  1-1e-15) x =  1;
      else if (x  < -1+1e-15) x = -1;
      y = st*sin(phi);          /* compute y avoiding roundoff errors */
      if (fabs(y) <    1e-15) y =  0;
      else if (y  >  1-1e-15) y =  1;
      else if (y  < -1+1e-15) y = -1;
      fprintf(out, "%g %g %g\n", x, y, z);
      phi += d;                 /* print the next point and */
    }                           /* increment the azimuth angle */
  }
  fprintf(out, "%g %g %g\n", 0.0, 0.0, 1.0);
  if (fclose(out) != 0)         /* close the output file and */
    error(E_FWRITE, fname);     /* check for a write error */
  printf("done.\n");            /* print a success message */

  return 0;                     /* return 'ok' */
}  /* main() */
